@@ -16,10 +16,13 @@ mod fallback;
 #[inline]
 pub fn atomic_is_lock_free<T>() -> bool {
     match mem::size_of::<T>() {
+        #[cfg(target_has_atomic = "8")]
         1 if mem::align_of::<T>() >= 1 => true,
+        #[cfg(target_has_atomic = "16")]
         2 if mem::align_of::<T>() >= 2 => true,
+        #[cfg(target_has_atomic = "32")]
         4 if mem::align_of::<T>() >= 4 => true,
-        #[cfg(any(target_pointer_width = "64", target_arch = "x86", target_arch = "arm"))]
+        #[cfg(target_has_atomic = "64")]
         8 if mem::align_of::<T>() >= 8 => true,
         _ => false,
     }
@@ -38,16 +41,19 @@ unsafe fn atomic_load_raw<T>(dst: *mut T, order: Ordering) -> T {
 #[inline]
 pub unsafe fn atomic_load<T>(dst: *mut T, order: Ordering) -> T {
     match mem::size_of::<T>() {
+        #[cfg(target_has_atomic = "8")]
         1 if mem::align_of::<T>() >= 1 => {
             mem::transmute_copy(&atomic_load_raw(dst as *mut u8, order))
         }
+        #[cfg(target_has_atomic = "16")]
         2 if mem::align_of::<T>() >= 2 => {
             mem::transmute_copy(&atomic_load_raw(dst as *mut u16, order))
         }
+        #[cfg(target_has_atomic = "32")]
         4 if mem::align_of::<T>() >= 4 => {
             mem::transmute_copy(&atomic_load_raw(dst as *mut u32, order))
         }
-        #[cfg(any(target_pointer_width = "64", target_arch = "x86", target_arch = "arm"))]
+        #[cfg(target_has_atomic = "64")]
         8 if mem::align_of::<T>() >= 8 => {
             mem::transmute_copy(&atomic_load_raw(dst as *mut u64, order))
         }
@@ -68,16 +74,19 @@ unsafe fn atomic_store_raw<T>(dst: *mut T, val: T, order: Ordering) {
 #[inline]
 pub unsafe fn atomic_store<T>(dst: *mut T, val: T, order: Ordering) {
     match mem::size_of::<T>() {
+        #[cfg(target_has_atomic = "8")]
         1 if mem::align_of::<T>() >= 1 => {
             atomic_store_raw(dst as *mut u8, mem::transmute_copy(&val), order)
         }
+        #[cfg(target_has_atomic = "16")]
         2 if mem::align_of::<T>() >= 2 => {
             atomic_store_raw(dst as *mut u16, mem::transmute_copy(&val), order)
         }
+        #[cfg(target_has_atomic = "32")]
         4 if mem::align_of::<T>() >= 4 => {
             atomic_store_raw(dst as *mut u32, mem::transmute_copy(&val), order)
         }
-        #[cfg(any(target_pointer_width = "64", target_arch = "x86", target_arch = "arm"))]
+        #[cfg(target_has_atomic = "64")]
         8 if mem::align_of::<T>() >= 8 => {
             atomic_store_raw(dst as *mut u64, mem::transmute_copy(&val), order)
         }
@@ -98,16 +107,19 @@ unsafe fn atomic_swap_raw<T>(dst: *mut T, val: T, order: Ordering) -> T {
 #[inline]
 pub unsafe fn atomic_swap<T>(dst: *mut T, val: T, order: Ordering) -> T {
     match mem::size_of::<T>() {
+        #[cfg(target_has_atomic = "8")]
         1 if mem::align_of::<T>() >= 1 => {
             mem::transmute_copy(&atomic_swap_raw(dst as *mut u8, mem::transmute_copy(&val), order))
         }
+        #[cfg(target_has_atomic = "16")]
         2 if mem::align_of::<T>() >= 2 => {
             mem::transmute_copy(&atomic_swap_raw(dst as *mut u16, mem::transmute_copy(&val), order))
         }
+        #[cfg(target_has_atomic = "32")]
         4 if mem::align_of::<T>() >= 4 => {
             mem::transmute_copy(&atomic_swap_raw(dst as *mut u32, mem::transmute_copy(&val), order))
         }
-        #[cfg(any(target_pointer_width = "64", target_arch = "x86", target_arch = "arm"))]
+        #[cfg(target_has_atomic = "64")]
         8 if mem::align_of::<T>() >= 8 => {
             mem::transmute_copy(&atomic_swap_raw(dst as *mut u64, mem::transmute_copy(&val), order))
         }
@@ -162,6 +174,7 @@ pub unsafe fn atomic_compare_exchange<T>(dst: *mut T,
                                          failure: Ordering)
                                          -> Result<T, T> {
     match mem::size_of::<T>() {
+        #[cfg(target_has_atomic = "8")]
         1 if mem::align_of::<T>() >= 1 => {
             mem::transmute_copy(&atomic_compare_exchange_raw(dst as *mut u8,
                                                              mem::transmute_copy(&current),
@@ -169,6 +182,7 @@ pub unsafe fn atomic_compare_exchange<T>(dst: *mut T,
                                                              success,
                                                              failure))
         }
+        #[cfg(target_has_atomic = "16")]
         2 if mem::align_of::<T>() >= 2 => {
             mem::transmute_copy(&atomic_compare_exchange_raw(dst as *mut u16,
                                                              mem::transmute_copy(&current),
@@ -176,6 +190,7 @@ pub unsafe fn atomic_compare_exchange<T>(dst: *mut T,
                                                              success,
                                                              failure))
         }
+        #[cfg(target_has_atomic = "32")]
         4 if mem::align_of::<T>() >= 4 => {
             mem::transmute_copy(&atomic_compare_exchange_raw(dst as *mut u32,
                                                              mem::transmute_copy(&current),
@@ -183,7 +198,7 @@ pub unsafe fn atomic_compare_exchange<T>(dst: *mut T,
                                                              success,
                                                              failure))
         }
-        #[cfg(any(target_pointer_width = "64", target_arch = "x86", target_arch = "arm"))]
+        #[cfg(target_has_atomic = "64")]
         8 if mem::align_of::<T>() >= 8 => {
             mem::transmute_copy(&atomic_compare_exchange_raw(dst as *mut u64,
                                                              mem::transmute_copy(&current),
@@ -248,6 +263,7 @@ pub unsafe fn atomic_compare_exchange_weak<T>(dst: *mut T,
                                               failure: Ordering)
                                               -> Result<T, T> {
     match mem::size_of::<T>() {
+        #[cfg(target_has_atomic = "8")]
         1 if mem::align_of::<T>() >= 1 => {
             mem::transmute_copy(&atomic_compare_exchange_weak_raw(dst as *mut u8,
                                                                   mem::transmute_copy(&current),
@@ -255,6 +271,7 @@ pub unsafe fn atomic_compare_exchange_weak<T>(dst: *mut T,
                                                                   success,
                                                                   failure))
         }
+        #[cfg(target_has_atomic = "16")]
         2 if mem::align_of::<T>() >= 2 => {
             mem::transmute_copy(&atomic_compare_exchange_weak_raw(dst as *mut u16,
                                                                   mem::transmute_copy(&current),
@@ -262,6 +279,7 @@ pub unsafe fn atomic_compare_exchange_weak<T>(dst: *mut T,
                                                                   success,
                                                                   failure))
         }
+        #[cfg(target_has_atomic = "32")]
         4 if mem::align_of::<T>() >= 4 => {
             mem::transmute_copy(&atomic_compare_exchange_weak_raw(dst as *mut u32,
                                                                   mem::transmute_copy(&current),
@@ -269,7 +287,7 @@ pub unsafe fn atomic_compare_exchange_weak<T>(dst: *mut T,
                                                                   success,
                                                                   failure))
         }
-        #[cfg(any(target_pointer_width = "64", target_arch = "x86", target_arch = "arm"))]
+        #[cfg(target_has_atomic = "64")]
         8 if mem::align_of::<T>() >= 8 => {
             mem::transmute_copy(&atomic_compare_exchange_weak_raw(dst as *mut u64,
                                                                   mem::transmute_copy(&current),
@@ -296,9 +314,22 @@ pub unsafe fn atomic_add<T: Copy>(dst: *mut T, val: T, order: Ordering) -> T
     where Wrapping<T>: ops::Add<Output = Wrapping<T>>
 {
     match mem::size_of::<T>() {
-        1 | 2 | 4 => atomic_add_raw(dst, val, order),
-        #[cfg(any(target_pointer_width = "64", target_arch = "x86", target_arch = "arm"))]
-        8 => atomic_add_raw(dst, val, order),
+        #[cfg(target_has_atomic = "8")]
+        1 if mem::align_of::<T>() >= 1 => {
+            mem::transmute_copy(&atomic_add_raw(dst as *mut u8, mem::transmute_copy(&val), order))
+        }
+        #[cfg(target_has_atomic = "16")]
+        2 if mem::align_of::<T>() >= 2 => {
+            mem::transmute_copy(&atomic_add_raw(dst as *mut u16, mem::transmute_copy(&val), order))
+        }
+        #[cfg(target_has_atomic = "32")]
+        4 if mem::align_of::<T>() >= 4 => {
+            mem::transmute_copy(&atomic_add_raw(dst as *mut u32, mem::transmute_copy(&val), order))
+        }
+        #[cfg(target_has_atomic = "64")]
+        8 if mem::align_of::<T>() >= 8 => {
+            mem::transmute_copy(&atomic_add_raw(dst as *mut u64, mem::transmute_copy(&val), order))
+        }
         _ => fallback::atomic_add(dst, val),
     }
 }
@@ -318,9 +349,22 @@ pub unsafe fn atomic_sub<T: Copy>(dst: *mut T, val: T, order: Ordering) -> T
     where Wrapping<T>: ops::Sub<Output = Wrapping<T>>
 {
     match mem::size_of::<T>() {
-        1 | 2 | 4 => atomic_sub_raw(dst, val, order),
-        #[cfg(any(target_pointer_width = "64", target_arch = "x86", target_arch = "arm"))]
-        8 => atomic_sub_raw(dst, val, order),
+        #[cfg(target_has_atomic = "8")]
+        1 if mem::align_of::<T>() >= 1 => {
+            mem::transmute_copy(&atomic_sub_raw(dst as *mut u8, mem::transmute_copy(&val), order))
+        }
+        #[cfg(target_has_atomic = "16")]
+        2 if mem::align_of::<T>() >= 2 => {
+            mem::transmute_copy(&atomic_sub_raw(dst as *mut u16, mem::transmute_copy(&val), order))
+        }
+        #[cfg(target_has_atomic = "32")]
+        4 if mem::align_of::<T>() >= 4 => {
+            mem::transmute_copy(&atomic_sub_raw(dst as *mut u32, mem::transmute_copy(&val), order))
+        }
+        #[cfg(target_has_atomic = "64")]
+        8 if mem::align_of::<T>() >= 8 => {
+            mem::transmute_copy(&atomic_sub_raw(dst as *mut u64, mem::transmute_copy(&val), order))
+        }
         _ => fallback::atomic_sub(dst, val),
     }
 }
@@ -341,9 +385,22 @@ pub unsafe fn atomic_and<T: Copy + ops::BitAnd<Output = T>>(dst: *mut T,
                                                             order: Ordering)
                                                             -> T {
     match mem::size_of::<T>() {
-        1 | 2 | 4 => atomic_and_raw(dst, val, order),
-        #[cfg(any(target_pointer_width = "64", target_arch = "x86", target_arch = "arm"))]
-        8 => atomic_and_raw(dst, val, order),
+        #[cfg(target_has_atomic = "8")]
+        1 if mem::align_of::<T>() >= 1 => {
+            mem::transmute_copy(&atomic_and_raw(dst as *mut u8, mem::transmute_copy(&val), order))
+        }
+        #[cfg(target_has_atomic = "16")]
+        2 if mem::align_of::<T>() >= 2 => {
+            mem::transmute_copy(&atomic_and_raw(dst as *mut u16, mem::transmute_copy(&val), order))
+        }
+        #[cfg(target_has_atomic = "32")]
+        4 if mem::align_of::<T>() >= 4 => {
+            mem::transmute_copy(&atomic_and_raw(dst as *mut u32, mem::transmute_copy(&val), order))
+        }
+        #[cfg(target_has_atomic = "64")]
+        8 if mem::align_of::<T>() >= 8 => {
+            mem::transmute_copy(&atomic_and_raw(dst as *mut u64, mem::transmute_copy(&val), order))
+        }
         _ => fallback::atomic_and(dst, val),
     }
 }
@@ -364,9 +421,22 @@ pub unsafe fn atomic_or<T: Copy + ops::BitOr<Output = T>>(dst: *mut T,
                                                           order: Ordering)
                                                           -> T {
     match mem::size_of::<T>() {
-        1 | 2 | 4 => atomic_or_raw(dst, val, order),
-        #[cfg(any(target_pointer_width = "64", target_arch = "x86", target_arch = "arm"))]
-        8 => atomic_or_raw(dst, val, order),
+        #[cfg(target_has_atomic = "8")]
+        1 if mem::align_of::<T>() >= 1 => {
+            mem::transmute_copy(&atomic_or_raw(dst as *mut u8, mem::transmute_copy(&val), order))
+        }
+        #[cfg(target_has_atomic = "16")]
+        2 if mem::align_of::<T>() >= 2 => {
+            mem::transmute_copy(&atomic_or_raw(dst as *mut u16, mem::transmute_copy(&val), order))
+        }
+        #[cfg(target_has_atomic = "32")]
+        4 if mem::align_of::<T>() >= 4 => {
+            mem::transmute_copy(&atomic_or_raw(dst as *mut u32, mem::transmute_copy(&val), order))
+        }
+        #[cfg(target_has_atomic = "64")]
+        8 if mem::align_of::<T>() >= 8 => {
+            mem::transmute_copy(&atomic_or_raw(dst as *mut u64, mem::transmute_copy(&val), order))
+        }
         _ => fallback::atomic_or(dst, val),
     }
 }
@@ -387,9 +457,22 @@ pub unsafe fn atomic_xor<T: Copy + ops::BitXor<Output = T>>(dst: *mut T,
                                                             order: Ordering)
                                                             -> T {
     match mem::size_of::<T>() {
-        1 | 2 | 4 => atomic_xor_raw(dst, val, order),
-        #[cfg(any(target_pointer_width = "64", target_arch = "x86", target_arch = "arm"))]
-        8 => atomic_xor_raw(dst, val, order),
+        #[cfg(target_has_atomic = "8")]
+        1 if mem::align_of::<T>() >= 1 => {
+            mem::transmute_copy(&atomic_xor_raw(dst as *mut u8, mem::transmute_copy(&val), order))
+        }
+        #[cfg(target_has_atomic = "16")]
+        2 if mem::align_of::<T>() >= 2 => {
+            mem::transmute_copy(&atomic_xor_raw(dst as *mut u16, mem::transmute_copy(&val), order))
+        }
+        #[cfg(target_has_atomic = "32")]
+        4 if mem::align_of::<T>() >= 4 => {
+            mem::transmute_copy(&atomic_xor_raw(dst as *mut u32, mem::transmute_copy(&val), order))
+        }
+        #[cfg(target_has_atomic = "64")]
+        8 if mem::align_of::<T>() >= 8 => {
+            mem::transmute_copy(&atomic_xor_raw(dst as *mut u64, mem::transmute_copy(&val), order))
+        }
         _ => fallback::atomic_xor(dst, val),
     }
 }
